import typer
from pathlib import Path

from parxy_cli.console.console import Console


app = typer.Typer()

console = Console()


@app.command()
def env():
    """Create an environment file with Parxy configuration."""
    from importlib.resources import files

    try:
        # Get the example file content from the package
        example_content = files('parxy_cli').joinpath('.env.example').read_text()

        env_file_path: Path = Path.cwd() / '.env'

        console.action('Create env file')

        # Check if .env already exists
        if env_file_path.exists():
            console.highlight('.env file already exists')
            console.newline()
            overwrite = typer.confirm('Do you want to overwrite it?', default=False)
            if not overwrite:
                console.faint('Leaving your file as is.')
                return

        # Write the content to .env
        env_file_path.write_text(example_content)

        console.success(
            '[success]Created .env file[/success] with default configuration.'
        )
        console.faint('Edit the file to configure your settings.')
    except Exception as e:
        console.error(f'Error creating .env file: {str(e)}')
        raise typer.Exit(1)



from abc import ABC
from enum import IntEnum
from typing import List, Optional, Any

from pydantic import BaseModel


class BoundingBox(BaseModel):
    x0: float
    y0: float
    x1: float
    y1: float


class Style(BaseModel):
    font_name: Optional[str] = None
    font_size: Optional[float] = None
    font_style: Optional[str] = None
    color: Optional[str] = None
    alpha: Optional[int] = None
    weight: Optional[float] = None


class Character(BaseModel):
    text: str
    bbox: Optional[BoundingBox] = None
    style: Optional[Style] = None
    page: Optional[int] = None
    source_data: Optional[dict[str, Any]] = None

    def isEmpty(self) -> bool:
        return not self.text or self.text.strip() == ''


class Span(BaseModel):
    text: str
    bbox: Optional[BoundingBox] = None
    style: Optional[Style] = None
    characters: Optional[List[Character]] = None
    page: Optional[int] = None
    source_data: Optional[dict[str, Any]] = None

    def isEmpty(self) -> bool:
        return not self.text or self.text.strip() == ''


class Line(BaseModel):
    text: str
    bbox: Optional[BoundingBox] = None
    style: Optional[Style] = None
    spans: Optional[List[Span]] = None
    page: Optional[int] = None
    source_data: Optional[dict[str, Any]] = None

    def isEmpty(self) -> bool:
        return not self.text or self.text.strip() == ''


class Block(BaseModel, ABC):
    type: str
    bbox: Optional[BoundingBox] = None
    page: Optional[int] = None
    source_data: Optional[dict[str, Any]] = None


class TextBlock(BaseModel):
    type: str
    bbox: Optional[BoundingBox] = None
    page: Optional[int] = None
    source_data: Optional[dict[str, Any]] = None
    category: Optional[str] = None
    style: Optional[Style] = None
    level: Optional[int] = None
    lines: Optional[List[Line]] = None
    text: str

    def isEmpty(self) -> bool:
        return not self.text or self.text.strip() == ''


class ImageBlock(Block): ...


class TableBlock(Block): ...


class Page(BaseModel):
    number: int
    width: Optional[float] = None
    height: Optional[float] = None
    blocks: Optional[List[TextBlock | ImageBlock | TableBlock]] = None
    text: str
    source_data: Optional[dict[str, Any]] = None

    def isEmpty(self) -> bool:
        return not self.text or self.text.strip() == ''


class Metadata(BaseModel):
    title: Optional[str] = None
    author: Optional[str] = None
    subject: Optional[str] = None
    keywords: Optional[str] = None
    creator: Optional[str] = None
    producer: Optional[str] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None


class Document(BaseModel):
    filename: Optional[str] = None
    language: Optional[str] = None
    metadata: Optional[Metadata] = None
    pages: List[Page]
    outline: Optional[List[str]] = None
    source_data: Optional[dict[str, Any]] = None
    parsing_metadata: Optional[dict[str, Any]] = None

    def isEmpty(self) -> bool:
        return all(page.isEmpty() for page in self.pages)

    def text(self, page_separator: str = '---') -> str:
        """Get the full text content of the document.

        Parameters
        ----------
        page_separator : str, optional
            String to use as separator between pages, by default "---"
            Set to empty string or None to disable page separation

        Returns
        -------
        str
            The concatenated text of all pages with optional separators
        """
        if not self.pages:
            return ''

        # Filter out empty pages
        texts = [page.text.strip() for page in self.pages if page.text]

        if not texts:
            return ''

        # Add separator between pages if specified
        if page_separator:
            return f'\n{page_separator}\n'.join(texts)

        return '\n'.join(texts)

    def markdown(self) -> str:
        """Get the document content formatted as Markdown.

        The method attempts to preserve the document structure by:
        1. Converting TextBlocks to paragraphs based on their category
        2. Preserving line breaks where meaningful
        3. Adding section headers based on block levels

        Returns
        -------
        str
            The document content formatted as Markdown
        """
        if not self.pages:
            return ''

        markdown_parts = []

        for page in self.pages:
            if not page.blocks:
                if page.text.strip():
                    markdown_parts.append(page.text.strip())
                continue

            page_parts = []

            for block in page.blocks:
                if isinstance(block, TextBlock):
                    # Handle different block categories
                    if block.category and block.category.lower() in [
                        'heading',
                        'title',
                        'header',
                    ]:
                        # Determine heading level (h1-h6) based on block level or default to h2
                        level = min(block.level or 2, 6)
                        page_parts.append(f'{"#" * level} {block.text.strip()}')
                    elif block.category and block.category.lower() == 'list':
                        # Convert to bullet points
                        for line in block.text.splitlines():
                            if line.strip():
                                page_parts.append(f'- {line.strip()}')
                    else:
                        # Regular paragraph
                        if block.text.strip():
                            page_parts.append(block.text.strip())

                elif isinstance(block, ImageBlock):
                    # Placeholder for images - could be enhanced with actual image data
                    page_parts.append('![Image]')

                elif isinstance(block, TableBlock):
                    # Placeholder for tables - could be enhanced with actual table data
                    page_parts.append('| Table content |')

            if page_parts:
                markdown_parts.append('\n\n'.join(page_parts))

        return '\n\n'.join(markdown_parts)


class HierarchyLevel(IntEnum):
    PAGE = 0
    PARAGRAPH = 1
    BLOCK = 2
    LINE = 3
    SPAN = 4
    WORD = 5
    CHARACTER = 6


def estimate_lines_from_block(
    block: TextBlock, default_font_size: float = 11
) -> TextBlock:
    """Estimate line-level layout inside a text block by splitting text and assigning bounding boxes.

    Args:
        block (TextBlock): Text block to estimate lines for.
        default_font_size (float): Default font size if not specified. Default to 11.

    Returns:
        TextBlock: The same block with its `lines` field populated.
    """
    if not block.text or not block.bbox or block.lines is not None:
        return block

    block.lines = []

    # Try to split by explicit newlines first
    raw_lines = block.text.splitlines()
    n_lines = len(raw_lines)
    # fallback: if no explicit \n but text is too long, you might want to wrap it — skipped here

    if n_lines == 0:
        raw_lines = [block.text]
        n_lines = 1

    # Estimate line height
    font_size = block.style.font_size if block.style else default_font_size
    line_height = font_size * 1.1  # 10% line spacing
    total_height = block.bbox.y1 - block.bbox.y0

    # If bbox is taller than sum of line heights, spread the lines proportionally
    if n_lines > 1:
        estimated_line_height = total_height / n_lines
    else:
        estimated_line_height = line_height

    x0 = block.bbox.x0
    x1 = block.bbox.x1
    y_top = block.bbox.y0

    for idx, line_text in enumerate(raw_lines):
        y0 = y_top + idx * estimated_line_height
        y1 = y0 + estimated_line_height
        line_bbox = BoundingBox(x0=x0, y0=y0, x1=x1, y1=y1)

        line = Line(
            text=line_text,
            bbox=line_bbox,
            style=block.style,
            page=block.page,
            source_data={'source': 'split_from_block'},
            spans=None,
        )
        block.lines.append(line)
    return block



from enum import Enum
from typing import Annotated
# from typer import Context, Option


class Level(str, Enum):
    """Valid extraction levels."""

    PAGE = 'page'
    BLOCK = 'block'
    LINE = 'line'
    SPAN = 'span'
    CHARACTER = 'character'


class OutputMode(str, Enum):
    """Valid output modes for parse command."""

    JSON = 'json'
    PLAIN = 'plain'
    MARKDOWN = 'markdown'



from pathlib import Path

import typer

from parxy_cli.console.console import Console

app = typer.Typer()

console = Console()


@app.command()
def docker():
    """Create a Docker Compose file to run self-hostable parsers (experimental)."""

    from importlib.resources import files

    # Get the example file content from the package
    try:
        example_content = (
            files('parxy_cli').joinpath('compose.example.yaml').read_text()
        )

        console.action('Create compose.yaml file')

        compose_file_path: Path = Path.cwd() / 'compose.yaml'

        # Check if compose.yaml already exists
        if compose_file_path.exists():
            console.highlight('compose.yaml file already exists')
            console.newline()
            overwrite = typer.confirm('Do you want to overwrite it?', default=False)
            if not overwrite:
                console.faint('Leaving your file as is.')
                return

        # Write the content to compose.yaml
        compose_file_path.write_text(example_content)

        console.print(
            '[success]Created compose.yaml[/success] file with default configuration.'
        )
        console.newline()
        console.markdown(
            'Execute `docker compose pull` and `docker compose up -d` to start the services.'
        )

    except Exception as e:
        console.error(f'Error creating compose.yaml file: {str(e)}')
        raise typer.Exit(1)



import typer

from parxy_core.facade import Parxy
from parxy_cli.console.console import Console

app = typer.Typer()

console = Console()


@app.command()
def drivers():
    """List supported drivers."""

    drivers = Parxy.drivers()

    console.action('Parxy available drivers')

    console.print(f'{len(drivers)} drivers supported:')

    driver_list = '\n'.join(f'- {driver}' for driver in drivers)
    console.markdown(driver_list)



"""Command line interface for Parxy."""

import os
import sys
from typing import Optional, List

import typer
from typing_extensions import Annotated
from importlib.metadata import version as metadata_version

from parxy_cli.console.console import Console
from parxy_cli.commands.docker import app as docker_command
from parxy_cli.commands.parse import app as parse_command
from parxy_cli.commands.preview import app as preview_command
from parxy_cli.commands.drivers import app as drivers_command
from parxy_cli.commands.env import app as env_command
from parxy_cli.commands.version import app as version_command
from parxy_cli.commands.markdown import app as markdown_command


# Create typer app
app = typer.Typer(
    name='parxy',
    help='Parxy document processing gateway.',
    no_args_is_help=True,
    add_completion=False,
    pretty_exceptions_enable=False,
)

# Create themed console
console = Console()


def version_callback(value: bool):
    if value:
        try:
            parxy_version = metadata_version('parxy')
        except Exception:
            parxy_version = 'Development version'

        console.print(f'[blue]▣[/blue] [bold]Parxy[/bold]')
        console.print(f'[faint][italic]Every document matters.[/italic][/faint]')
        console.newline()
        console.print(f'Version: {parxy_version}')
        console.newline()
        console.markdown('For more information on Parxy run `parxy version`.')
        raise typer.Exit()


@app.callback()
def main(
    version: Annotated[
        Optional[bool],
        typer.Option(
            '--version',
            callback=version_callback,
            is_eager=True,
            help='Show Parxy version',
        ),
    ] = None,
):
    """Define the common command options"""

    pass


app.add_typer(docker_command)
app.add_typer(parse_command)
app.add_typer(preview_command)
app.add_typer(drivers_command)
app.add_typer(env_command)
app.add_typer(version_command)
app.add_typer(markdown_command)


def main():
    """Entry point for the CLI."""
    app()



import io
from pathlib import Path
from typing import TYPE_CHECKING

from parxy_core.exceptions.authentication_exception import AuthenticationException

# Type hints that will be available at runtime when unstructured is installed
if TYPE_CHECKING:
    from landingai_ade.types import ParseResponse
    from landingai_ade.types.parse_response import ChunkGroundingBox
else:
    # Placeholder types for when package is not installed
    ParseResponse = object
    ChunkGroundingBox = object

from parxy_core.drivers import Driver
from parxy_core.models import Document, Metadata, TextBlock, Page, BoundingBox


class LandingAIADEDriver(Driver):
    def _initialize_driver(self):
        try:
            from landingai_ade import LandingAIADE
        except ImportError as e:
            raise ImportError(
                'LandingAI dependencies not installed. '
                "Install with 'pip install parxy[landingai]'"
            ) from e

        self.__client = LandingAIADE(
            apikey=self._config.api_key.get_secret_value()
            if self._config and self._config.api_key
            else None,
            **self._config.model_dump() if self._config else {},
        )

    def _handle(
        self,
        file: str | io.BytesIO | bytes,
        level: str = 'page',
        **kwargs,
    ) -> Document:
        from landingai_ade import AuthenticationError

        try:
            parse_response = self.__client.parse(document=Path(file), **kwargs)

        except AuthenticationError as aex:
            raise AuthenticationException(
                message=str(aex),
                service=self.__class__,
            ) from aex

        return landingaiade_to_parxy(parse_response)


def landingaiade_to_parxy(parsed_data: ParseResponse) -> Document:
    # Group chunks by page
    page_chunks = {}
    for chunk in parsed_data.chunks:
        # Get the first grounding location (chunks can span multiple locations)
        if chunk.grounding:
            page_num = chunk.grounding.page
            if page_num not in page_chunks:
                page_chunks[page_num] = []
            page_chunks[page_num].append(chunk)

    # Convert to pages
    pages = []
    for page_num in sorted(page_chunks.keys()):
        chunks = page_chunks[page_num]
        blocks = []
        page_text_parts = []

        for chunk in chunks:
            chunk_text = chunk.markdown

            page_text_parts.append(chunk_text)
            chunk_type = chunk.type

            # Get bounding box from first grounding
            bbox = None
            page = None
            if chunk.grounding:
                grounding = chunk.grounding
                bbox = _convert_bbox(grounding.box)
                page = grounding.page

            # Create the appropriate block type
            block = TextBlock(
                type='text',
                bbox=bbox,
                page=page,
                category=chunk_type,
                text=chunk_text,
                source_data=chunk.model_dump(),
            )

            blocks.append(block)

        # Create page
        page_text = '\n'.join(page_text_parts)
        page = Page(
            number=page_num,
            blocks=blocks,
            text=page_text,
            source_data={'page_index': page_num},
        )
        pages.append(page)

    # Create the document
    document = Document(
        filename=parsed_data.metadata.filename,
        pages=pages,
        source_data=parsed_data.metadata.model_dump(),
    )

    return document


def _convert_bbox(box: ChunkGroundingBox) -> BoundingBox:
    """Convert box coordinates from l,t,r,b format to x0,y0,x1,y1 format."""
    return BoundingBox(x0=box.left, y0=box.top, x1=box.right, y1=box.bottom)



"""Command line interface for Parxy document processing."""

import sys
import platform
from importlib.metadata import version as metadata_version
import typer

from parxy_cli.console.console import Console


app = typer.Typer()

console = Console()


@app.command()
def version():
    """Print Parxy version information."""
    try:
        parxy_version = metadata_version('parxy')
    except Exception:
        parxy_version = 'Development version'

    console.parxy()

    console.print(f'Version: {parxy_version}')

    console.newline()
    console.muted(
        f'Python {sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}'
    )
    console.muted(f'Platform: {platform.platform()}')



import io
from typing import TYPE_CHECKING

from parxy_core.models.config import LlamaParseConfig

# Type hints that will be available at runtime when llama_cloud_services is installed
if TYPE_CHECKING:
    from llama_cloud_services import LlamaParse
    from llama_cloud_services.parse.types import JobResult, PageItem, Page as LlamaPage
else:
    # Placeholder types for when package is not installed
    LlamaParse = None
    JobResult = object
    PageItem = object
    LlamaPage = object

from parxy_core.drivers import Driver
from parxy_core.models import Document, Page, BoundingBox, TextBlock, HierarchyLevel
from parxy_core.exceptions import (
    ParsingException,
    AuthenticationException,
    FileNotFoundException,
)


class LlamaParseDriver(Driver):
    """Llama Cloud Services document processing via LlamaParse API.

    This parser interacts with the LlamaParse cloud service to extract document text,
    supporting page- and block-level extraction.

    Attributes
    ----------
    supported_levels : list of str
        The supported extraction levels: `page`, `block`.
    client : LlamaParse
        The LlamaParse client instance.
    """

    _config: LlamaParseConfig

    # supported_levels: list[str] = ["page", "block"]

    def _initialize_driver(self):
        """Initialize the Llama Parse driver.

        Raises
        ------
        ImportError
            If LlamaParse dependencies are not installed
        """
        try:
            from llama_cloud_services import LlamaParse
        except ImportError as e:
            raise ImportError(
                'LlamaParse dependencies not installed. '
                "Install with 'pip install parxy[llama]'"
            ) from e

        self.__client = LlamaParse(
            api_key=self._config.api_key.get_secret_value()
            if self._config and self._config.api_key
            else None,
            **self._config.model_dump() if self._config else {},
        )

    def _handle(
        self,
        file: str | io.BytesIO | bytes,
        level: str = 'block',
        **kwargs,
    ) -> Document:
        """Parse a document using LlamaParse.

        Parameters
        ----
        file : str | io.BytesIO | bytes
            Path, URL or stream of the file to parse.
        level : str, optional
            Desired extraction level. Must be one of `supported_levels`. Default is `"block"`.
        raw : bool, optional
            If True, return the raw `JobResult` object from LlamaParse instead of a `Document`. Default is False.

        Returns
        -------
        Document
            A parsed `Document` in unified format.

        Raises
        ------
        ImportError
            If LlamaParse dependencies are not installed
        AuthenticationException
            If authentication with LlamaParse fails
        FileNotFoundException
            If the input file cannot be accessed
        ParsingException
            If any other parsing error occurs
        """

        extra_info = None
        if isinstance(file, (io.BytesIO, bytes)):
            extra_info = {'file_name': 'file-name'}

        try:
            res = self.__client.parse(file, extra_info=extra_info)
        except FileNotFoundError as fex:
            raise FileNotFoundException(fex, self.__class__) from fex
        except Exception as ex:
            # Handle HTTP status errors specifically for authentication failures
            if hasattr(ex, '__cause__') and ex.__cause__ is not None:
                cause = ex.__cause__
                if hasattr(cause, 'response') and cause.response is not None:
                    status_code = cause.response.status_code
                    error_detail = (
                        cause.response.json() if hasattr(cause.response, 'json') else {}
                    )

                    if status_code in (401, 403):
                        raise AuthenticationException(
                            message=str(
                                error_detail.get('detail', 'Authentication failed')
                            ),
                            service=self.__class__,
                            details={
                                'status_code': status_code,
                                'error_response': error_detail,
                            },
                        ) from ex

            # For all other errors, raise as parsing exception
            raise ParsingException(str(ex), self.__class__) from ex

        if res.error is not None:
            raise ParsingException(
                res.error, self.__class__, res.model_dump(exclude={'file_name'})
            )

        return llamaparse_to_parxy(doc=res, level=level)


def llamaparse_to_parxy(
    doc: JobResult,
    level: str,
) -> Document:
    """Convert a LlamaParse `JobResult` to a `Document` object.

    Parameters
    ----
    doc : JobResult
        The LlamaParse result object.
    level : str
        Desired extraction level.

    Returns
    -------
    Document
        The converted `Document` in unified format.
    """
    pages = [_convert_page(page, level.upper()) for page in doc.pages]
    return Document(
        filename=doc.file_name,
        pages=pages,
        source_data=doc.model_dump(exclude={'file_name', 'pages'}),
    )


def _convert_text_block(text_block: PageItem, page_number: int) -> TextBlock:
    """Convert a LlamaParse `PageItem` to a `TextBlock`.

    Parameters
    ----
    text_block : PageItem
        The LlamaParse page item.
    page_number : int
        The page number (0-based).

    Returns
    -------
    TextBlock
        The converted `TextBlock` object.
    """
    bbox = BoundingBox(
        x0=text_block.bBox.x,
        y0=text_block.bBox.y,
        x1=text_block.bBox.x + text_block.bBox.w,
        y1=text_block.bBox.y + text_block.bBox.h,
    )
    return TextBlock(
        type='text',
        category=text_block.type,
        level=text_block.lvl,
        text=text_block.value if text_block.value else '',
        bbox=bbox,
        page=page_number,
        source_data=text_block.model_dump(exclude={'bBox', 'value', 'type', 'lvl'}),
    )


def _convert_page(
    page: LlamaPage,
    level: str,
) -> Page:
    """Convert a LlamaParse `Page` to a `Page` object.

    Parameters
    ----
    page : LlamaPage
        The LlamaParse page object.
    level : str
        Desired extraction level.

    Returns
    -------
    Page
        The converted `Page` object.
    """
    text_blocks = None
    if HierarchyLevel[level] >= HierarchyLevel.BLOCK:
        text_blocks = [_convert_text_block(item, page.page - 1) for item in page.items]
    return Page(
        number=page.page - 1,
        width=page.width,
        height=page.height,
        text=page.text,
        blocks=text_blocks,
        source_data=page.model_dump(
            exclude={'page', 'text', 'items', 'width', 'height'}
        ),
    )



import io

import validators

from typing import TYPE_CHECKING

# Type hints that will be available at runtime when unstructured is installed
if TYPE_CHECKING:
    from unstructured.documents.elements import Element as UnstructuredElement
    from unstructured.documents.elements import Text as UnstructuredText
else:
    # Placeholder types for when package is not installed
    UnstructuredElement = object
    UnstructuredText = object


from parxy_core.drivers import Driver
from parxy_core.exceptions import FileNotFoundException, ParsingException
from parxy_core.models import Document, TextBlock, BoundingBox, Page, HierarchyLevel
from parxy_core.models.config import UnstructuredLocalConfig


class UnstructuredLocalDriver(Driver):
    """Parser implementation using the `unstructured` library.

    This parser reads a document using Unstructured partitioning and converts it into the unified `Document` format,
    supporting `page` and `paragraph` levels of detail.

    Attributes
    ----------
    supported_levels : list of str
        Supported extraction levels: `page`, `paragraph`.
    """

    supported_levels: list[str] = ['page', 'block']

    _config: UnstructuredLocalConfig

    def _initialize_driver(self):
        """Initialize the Unstructured driver."""

        try:
            from unstructured.partition.auto import partition

            self.__client = partition
        except ImportError as e:
            raise ImportError(
                'Unstructured dependencies not installed. '
                "Install with 'pip install parxy[unstructured_local]'"
            ) from e

    def _handle(
        self,
        file: str | io.BytesIO | bytes,
        level: str = 'paragraph',
        **kwargs,
    ) -> Document:
        """Parse a document using `unstructured` into the unified `Document` format.

        Parameters
        ----------
        file : str | io.BytesIO | bytes
            Path, URL or stream of the file to parse.
        level : str, optional
            Desired extraction level: `page` or `paragraph`. Default is `"paragraph"`.
        raw : bool, optional
            If True, returns the raw output of `unstructured.partition.auto.partition`. Default is False.
        **kwargs : dict
            Additional keyword arguments passed to `unstructured.partition.auto.partition`.

        Returns
        -------
        Document or list of Element
            The parsed document or the raw list of `unstructured` elements.
        """

        if level == 'block':
            level = 'paragraph'

        url, stream = None, None
        if isinstance(file, str):
            if validators.url(file) is True:
                url = file
                file = None
        elif isinstance(file, io.BytesIO):
            stream = file
            file = ''
        elif isinstance(file, bytes):
            stream = io.BytesIO(file)
            file = ''

        try:
            res = self.__client(
                filename=file, url=url, file=stream, languages=['eng'], **kwargs
            )
            # languages=["eng"] is added to suppress the warning about no language detected that is printed to stdout https://github.com/Unstructured-IO/unstructured/blob/8fd07fd9f6702d0b3536ca52db8fb017def252b4/unstructured/partition/pdf.py#L332
        except FileNotFoundError as fex:
            raise FileNotFoundException(fex, self.__class__) from fex
        except Exception as wex:
            if isinstance(wex, FileNotFoundException):
                raise wex
            raise ParsingException(
                wex.error_message if hasattr(wex, 'error_message') else str(wex),
                self.__class__,
                details=wex.value,
            ) from wex

        return unstructured_to_parxy(doc=res, level=level)


def unstructured_to_parxy(
    doc: list[UnstructuredElement],
    level: str,
) -> Document:
    """Convert a list of `unstructured` Elements to a unified `Document`.

    Parameters
    ----------
    doc : list of Element
        The list of elements returned by `unstructured.partition.auto.partition`.
    level : str
        Desired extraction level: `page` or `paragraph`.

    Returns
    -------
    Document
        The converted document.
    """
    level = level.upper()
    pages = [
        Page(
            number=i,
            text='',
            blocks=[] if HierarchyLevel[level] >= HierarchyLevel.PARAGRAPH else None,
        )
        for i in range(doc[-1].metadata.page_number)
    ]
    for el in doc:
        if not isinstance(el, UnstructuredText):
            continue  # Not supported yet
        if HierarchyLevel[level] >= HierarchyLevel.PARAGRAPH:
            parxy_element = _convert_text_block(el)
            pages[parxy_element.page].blocks.append(parxy_element)
        current_page = el.metadata.page_number - 1
        if pages[current_page].text != '':
            pages[current_page].text += '\n'
        pages[current_page].text += el.text
    return Document(
        filename=doc[0].metadata.filename,
        language=doc[0].metadata.languages[0] if doc[0].metadata.languages else None,
        pages=pages,
    )


def _convert_text_block(
    text_block: UnstructuredText,
) -> TextBlock:
    """Convert an `unstructured` Text element to a `TextBlock`.

    Parameters
    ----------
    text_block : Text
        The `unstructured` Text element.

    Returns
    -------
    TextBlock
        The converted text block.
    """
    x0, y0 = text_block.metadata.coordinates.points[0]
    x2, y2 = text_block.metadata.coordinates.points[2]
    bbox = BoundingBox(x0=x0, y0=y0, x1=x2 - x0, y1=y2 - y0)
    return TextBlock(
        type='text',
        bbox=bbox,
        category=text_block.category,
        text=text_block.text,
        page=text_block.metadata.page_number - 1,
        source_data=text_block.metadata.to_dict(),
    )



import logging

from typing import Dict, Optional, Callable, Self, List

from parxy_core.drivers.abstract_driver import Driver
from parxy_core.drivers.landingai import LandingAIADEDriver
from parxy_core.drivers.pymupdf import PyMuPdfDriver
from parxy_core.drivers.pdfact import PdfActDriver
from parxy_core.drivers.llamaparse import LlamaParseDriver
from parxy_core.drivers.llmwhisperer import LlmWhispererDriver
from parxy_core.drivers.unstructured_local import UnstructuredLocalDriver
from parxy_core.models import (
    PdfActConfig,
    LandingAIConfig,
    LlamaParseConfig,
    LlmWhispererConfig,
    UnstructuredLocalConfig,
    ParxyConfig,
)
from parxy_core.logging import create_isolated_logger


class DriverFactory:
    """Factory class for managing document parser drivers.

    This factory manages the registration and instantiation of parser drivers.
    It supports both built-in drivers and custom driver registration.
    Each driver can be configured with specific settings during instantiation.

    This is a singleton class - only one instance will ever exist.
    Use DriverFactory.build() to get the instance.

    Example
    -------
    >>> factory = DriverFactory.build()
    >>> driver = factory.driver('pymupdf')
    """

    # Private class variable to hold the DriverFactory instance
    __instance: Optional['DriverFactory'] = None

    __drivers: Dict[str, Driver] = {}
    """The created drivers"""

    __custom_creators: Dict[str, Callable[[], Driver]] = {}
    """The custom drivers"""

    _config: Optional[ParxyConfig] = None

    _logger: logging.Logger = None

    def __init__(self):
        raise Exception('Use `DriverFactory.build()` to create an instance.')

    @classmethod
    def build(cls) -> 'DriverFactory':
        """Create a new factory instance.

        Returns
        -------
        DriverFactory
            The singleton instance of the factory.
        """
        if cls.__instance is None:
            cls.__instance = cls.__new__(cls).initialize(ParxyConfig())
        return cls.__instance

    @classmethod
    def reset(cls):
        cls.__instance = None

    def initialize(self, config: ParxyConfig) -> Self:
        self._config = config

        self._logger = create_isolated_logger(
            name='parxy',
            level=self._config.logging_level,
            add_console_handler=True,
            add_file_handler=True if self._config.logging_file is not None else False,
            file_path=self._config.logging_file,
        )

        return self

    def driver(self, name: str = None) -> Driver:
        """Get a driver instance.

        Parameters
        ----------
        name : str
            The name of the registered driver to instantiate

        Returns
        -------
        Driver
            A new instance of the requested driver

        Raises
        ------
        ValueError
            If driver is not registered
        """

        if name is None:
            name = self.default_driver_name()

        if name not in self.__drivers:
            self.__drivers[name] = self.__create_driver(name)

        return self.__drivers.get(name)

    def default_driver_name(self) -> str:
        return self._config.default_driver

    def __create_driver(self, name: str) -> Driver:
        """Create a new driver instance.

        Parameters
        ----------
        name : str
            The name of the registered driver to instantiate

        Returns
        -------
        Driver
            A new instance of the requested driver

        Raises
        ------
        ValueError
            If driver is not registered
        """

        if name in self.__custom_creators:
            return self.__custom_creators[name]()

        method_name = f'_create_{name}_driver'

        if hasattr(self, method_name):
            return getattr(self, method_name)()

        raise ValueError(f'Driver [{name}] not supported.')

    def _create_pymupdf_driver(self) -> PyMuPdfDriver:
        """Create a PyMuPDF Driver instance.

        Returns
        -------
        PyMuPdfDriver
            A new instance
        """
        return PyMuPdfDriver(logger=self._logger)

    def _create_pdfact_driver(self) -> PdfActDriver:
        """Create a PdfAct Driver instance.

        Returns
        -------
        PdfActDriver
            A new instance
        """
        return PdfActDriver(config=PdfActConfig(), logger=self._logger)

    def _create_llamaparse_driver(self) -> LlamaParseDriver:
        """Create a LlamaParse Driver instance.

        Returns
        -------
        LlamaParseDriver
            A new instance
        """
        return LlamaParseDriver(
            config=LlamaParseConfig(),
            logger=self._logger,
        )

    def _create_llmwhisperer_driver(self) -> LlmWhispererDriver:
        """Create a LlmWhisperer Driver instance.

        Returns
        -------
        LlmWhispererDriver
            A new instance
        """
        return LlmWhispererDriver(
            config=LlmWhispererConfig(),
            logger=self._logger,
        )

    def _create_unstructured_local_driver(self) -> UnstructuredLocalDriver:
        """Create a Unstructured library (local installation via Python) Driver instance.

        Returns
        -------
        UnstructuredLocalDriver
            A new instance
        """
        return UnstructuredLocalDriver(
            config=UnstructuredLocalConfig(),
            logger=self._logger,
        )

    def _create_landingai_driver(self) -> LandingAIADEDriver:
        """Create a LandingAI ADE Driver instance.

        Returns
        -------
        LandingAIADEDriver
            A new instance
        """
        return LandingAIADEDriver(
            config=LandingAIConfig(),
            logger=self._logger,
        )

    def extend(self, name: str, callback: Callable[[], Driver]) -> 'DriverFactory':
        """Register a custom driver creator callable.

        Parameters
        ----------
        name : str
            The driver name
        callback : callable
            The function that creates the instance of the driver

        Raises
        ------
        ValueError
            If name is already registered
        """
        if name in self.__custom_creators:
            raise ValueError(f'Driver [{name}] already registered.')

        # TODO: pass logger to callback
        self.__custom_creators[name] = callback

        return self

    def get_drivers(self) -> Dict[str, Driver]:
        """Get all of the created "drivers".


        Returns
        -------
        Dict[str, Driver]
            The created driver instances
        """
        return self.__drivers

    def get_supported_drivers(self) -> List[str]:
        """Get the list of supported drivers.


        Returns
        -------
        List[str]
            The supported driver names
        """

        supported_drivers: List[str] = [
            'pymupdf',
            'pdfact',
            'landingai',
            'llamaparse',
            'llmwhisperer',
            'unstructured_local',
        ]

        return supported_drivers

    def forget_drivers(self) -> 'DriverFactory':
        """Forget all instantiated and custom "drivers".


        Returns
        -------
        Dict[str, Driver]
            The created driver instances
        """

        self.__drivers = {}

        self.__custom_creators = {}

        return self



"""Preview command for interactive document viewing."""

from typing import Optional, Annotated, List

import typer
from rich.layout import Layout
from rich.text import Text
from rich.panel import Panel
from rich.markdown import Markdown

from parxy_core.facade import Parxy
from parxy_core.models.models import Document, TextBlock, Metadata

from parxy_cli.models import Level
from parxy_cli.console.console import Console

app = typer.Typer()

console = Console()


def extract_toc(doc: Document) -> List[dict]:
    """
    Extract table of contents from document headings.

    Args:
        doc: The parsed document

    Returns:
        List of dicts with keys: 'text', 'level', 'page'
    """
    toc = []
    for page in doc.pages:
        if page.blocks:
            for block in page.blocks:
                if isinstance(block, TextBlock) and block.category in [
                    'heading',
                    'title',
                    'header',
                ]:
                    toc.append(
                        {
                            'text': block.text.strip(),
                            'level': block.level or 1,
                            'page': page.number,
                        }
                    )
    return toc


def format_metadata(metadata: Metadata, page_count: int) -> str:
    """
    Format document metadata for display.

    Args:
        metadata: Document metadata object
        page_count: Number of pages in document

    Returns:
        Formatted metadata string
    """

    lines = []

    if metadata is not None and metadata.title:
        lines.append(f'[bold]Title:[/bold] {metadata.title}')
    if metadata is not None and metadata.author:
        lines.append(f'[bold]Author:[/bold] {metadata.author}')
    if metadata is not None and metadata.subject:
        lines.append(f'[bold]Subject:[/bold] {metadata.subject}')
    if metadata is not None and metadata.keywords:
        lines.append(f'[bold]Keywords:[/bold] {metadata.keywords}')
    if metadata is not None and metadata.creator:
        lines.append(f'[bold]Creator:[/bold] {metadata.creator}')
    if metadata is not None and metadata.producer:
        lines.append(f'[bold]Producer:[/bold] {metadata.producer}')
    if metadata is not None and metadata.created_at:
        lines.append(f'[bold]Created:[/bold] {metadata.created_at}')
    if metadata is not None and metadata.updated_at:
        lines.append(f'[bold]Updated:[/bold] {metadata.updated_at}')

    lines.append(f'[bold]Pages:[/bold] {page_count}')

    return '\n'.join(lines) if lines else '[dim]No metadata available[/dim]'


def render_viewer_mode(doc: Document) -> Layout:
    """
    Render document in viewer mode with three panels: metadata, TOC, and content.

    Args:
        doc: The parsed document

    Returns:
        Layout object with three-panel view
    """
    # Extract TOC
    toc = extract_toc(doc)

    # Format metadata
    metadata_text = format_metadata(doc.metadata, len(doc.pages))

    # Format TOC
    if toc:
        toc_lines = []
        for item in toc:
            indent = '  ' * (item['level'] - 1)
            toc_lines.append(f'{indent}• {item["text"]} [dim](p.{item["page"]})[/dim]')
        toc_text = '\n'.join(toc_lines)
    else:
        toc_text = '[dim]No headings found[/dim]'

    # Create markdown content
    markdown_content = doc.markdown()

    # Create three panels
    metadata_panel = Panel(
        Text.from_markup(metadata_text),
        padding=(0, 1),
    )

    toc_panel = Panel(
        Text.from_markup(toc_text),
        title='Contents',
        border_style='blue',
        title_align='left',
        padding=(1, 2),
    )

    content_panel = Panel(
        Markdown(markdown_content, code_theme='monokai'),
        title='Document preview in markdown',
        title_align='left',
        padding=(1, 2),
    )

    # Use Layout to create three-column view
    layout = Layout()

    layout.split_column(
        Layout(metadata_panel, name='header', minimum_size=3),
        Layout(name='body', ratio=2),
    )

    layout['body'].split_row(
        Layout(toc_panel, name='toc', ratio=1),
        Layout(content_panel, name='content', ratio=2),
    )

    return layout


@app.command()
def preview(
    file: Annotated[
        str,
        typer.Argument(
            help='File to preview',
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
        ),
    ],
    driver: Annotated[
        Optional[str],
        typer.Option(
            '--driver',
            '-d',
            help='Driver to use for parsing (default: pymupdf or PARXY_DEFAULT_DRIVER)',
        ),
    ] = None,
    level: Annotated[
        Level,
        typer.Option(
            '--level',
            '-l',
            help='Extraction level',
        ),
    ] = Level.BLOCK,
    env_file: Annotated[
        str,
        typer.Option(
            '--env',
            '-e',
            help='Path to .env file with configuration',
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
        ),
    ] = '.env',
):
    """
    Preview a document in an interactive viewer with metadata, table of contents, and content.

    The preview is displayed in a scrollable three-panel layout showing:
    - Document metadata (title, author, pages, etc.)
    - Table of contents extracted from headings
    - Document content rendered as markdown
    """

    with console.shimmer(f'Processing {file} using {driver}...'):
        # Parse the document
        doc = Parxy.parse(
            file=file,
            level=level.value,
            driver_name=driver,
        )

    # Render the viewer layout
    layout = render_viewer_mode(doc)

    # Use Rich's pager for scrollable output with styles preserved
    console.print(layout)
    # with console.pager(styles=True):



"""Command line interface for Parxy document processing."""

from datetime import timedelta
from pathlib import Path
from typing import Optional, List, Annotated

import typer

from parxy_core.facade import Parxy
from parxy_core.models.models import Document

from parxy_cli.models import Level, OutputMode
from parxy_cli.console.console import Console

app = typer.Typer()

console = Console()


def collect_files_with_depth(
    directory: Path, pattern: str, max_depth: int, current_depth: int = 0
) -> List[Path]:
    """
    Recursively collect files matching pattern up to max_depth.

    Args:
        directory: Directory to search in
        pattern: File pattern to match (e.g., '*.pdf')
        max_depth: Maximum depth to recurse (0 = no recursion)
        current_depth: Current recursion depth (used internally)

    Returns:
        List of Path objects matching the pattern
    """
    files = []

    # Collect files in current directory
    files.extend(directory.glob(pattern))

    # Recurse into subdirectories if we haven't reached max depth
    if current_depth < max_depth:
        for subdir in directory.iterdir():
            if subdir.is_dir():
                files.extend(
                    collect_files_with_depth(
                        subdir, pattern, max_depth, current_depth + 1
                    )
                )

    return files


def collect_files(
    inputs: List[str], recursive: bool = False, max_depth: Optional[int] = None
) -> List[Path]:
    """
    Collect all files from the input list (files and/or folders).

    Args:
        inputs: List of file paths and/or folder paths
        recursive: Whether to search subdirectories
        max_depth: Maximum depth to recurse (only applies if recursive=True, None=unlimited)

    Returns:
        List of Path objects for all PDF files found
    """
    files = []

    for input_path in inputs:
        path = Path(input_path)

        if path.is_file():
            files.append(path)
        elif path.is_dir():
            if recursive:
                if max_depth is not None:
                    # Use depth-limited recursion
                    files.extend(collect_files_with_depth(path, '*.pdf', max_depth))
                else:
                    # Use unlimited recursion
                    files.extend(path.rglob('*.pdf'))
            else:
                # Non-recursive: only files in the given directory
                files.extend(path.glob('*.pdf'))
        else:
            console.warning(f'Path not found: {input_path}')

    return files


def get_output_extension(mode: OutputMode) -> str:
    """Get file extension based on output mode."""
    return {
        OutputMode.JSON: '.json',
        OutputMode.PLAIN: '.txt',
        OutputMode.MARKDOWN: '.md',
    }[mode]


def get_content(doc: Document, mode: OutputMode) -> str:
    """Extract content from document based on output mode."""
    if mode == OutputMode.JSON:
        return doc.model_dump_json(indent=2)
    elif mode == OutputMode.MARKDOWN:
        return doc.markdown()
    else:  # OutputMode.PLAIN
        return doc.text()


def process_file_with_driver(
    file_path: Path,
    driver: str,
    level: Level,
    mode: OutputMode,
    output_dir: Optional[Path],
    show: bool,
    use_driver_suffix: bool = False,
) -> tuple[str, int]:
    """
    Process a single file with a single driver.

    Args:
        file_path: Path to file to process
        driver: Driver name to use
        level: Extraction level
        mode: Output mode
        output_dir: Optional output directory
        show: Whether to show content in console
        use_driver_suffix: Whether to append driver name to output filename

    Returns:
        Tuple of (output_path, page_count)
    """
    # Parse the document
    doc = Parxy.parse(
        file=str(file_path),
        level=level.value,
        driver_name=driver,
    )

    # Get content
    content = get_content(doc, mode)

    # Determine output path
    if output_dir:
        output_dir.mkdir(parents=True, exist_ok=True)
        base_name = file_path.stem
    else:
        # Save in same directory as source file
        output_dir = file_path.parent
        base_name = file_path.stem

    # If multiple drivers, append driver name to filename
    if use_driver_suffix and driver:
        base_name = f'{base_name}-{driver}'

    extension = get_output_extension(mode)
    output_path = output_dir / f'{base_name}{extension}'

    # Save to file
    output_path.write_text(content, encoding='utf-8')

    # Show in console if requested
    if show:
        console.print(content)
        console.newline()

    return str(output_path), len(doc.pages)


def format_timedelta(td):
    days = td.days
    milliseconds = td.microseconds // 1000
    hours, remainder = divmod(td.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    parts = []
    if days > 0:
        parts.append(f'{days} day{"s" if days > 1 else ""}')
    if hours > 0:
        parts.append(f'{hours} hour{"s" if hours > 1 else ""}')
    if minutes > 0:
        parts.append(f'{minutes} min')
    if seconds > 0:
        parts.append(f'{seconds} sec')
    if milliseconds > 0:
        parts.append(f'{milliseconds} msec')

    return ', '.join(parts)


@app.command()
def parse(
    inputs: Annotated[
        List[str],
        typer.Argument(
            help='One or more files or folders to parse. Use --recursive to search subdirectories.',
        ),
    ],
    drivers: Annotated[
        Optional[List[str]],
        typer.Option(
            '--driver',
            '-d',
            help='Driver(s) to use for parsing. Can be specified multiple times. (default: pymupdf or PARXY_DEFAULT_DRIVER)',
        ),
    ] = None,
    level: Annotated[
        Level,
        typer.Option(
            '--level',
            '-l',
            help='Extraction level',
        ),
    ] = Level.PAGE,
    mode: Annotated[
        OutputMode,
        typer.Option(
            '--mode',
            '-m',
            help='Output mode: json (JSON serialization), plain (plain text), or markdown (markdown format)',
        ),
    ] = OutputMode.JSON,
    output_dir: Annotated[
        Optional[str],
        typer.Option(
            '--output',
            '-o',
            help='Directory to save output files. If not specified, files will be saved in the same directory as the source files.',
        ),
    ] = None,
    show: Annotated[
        bool,
        typer.Option(
            '--show',
            '-s',
            help='Show document content in console in addition to saving to files',
        ),
    ] = False,
    recursive: Annotated[
        bool,
        typer.Option(
            '--recursive',
            '-r',
            help='Recursively search subdirectories when processing folders',
        ),
    ] = False,
    max_depth: Annotated[
        Optional[int],
        typer.Option(
            '--max-depth',
            help='Maximum depth to recurse into subdirectories (only applies with --recursive). 0 = current directory only, 1 = one level down, etc.',
            min=0,
        ),
    ] = None,
    stop_on_failure: Annotated[
        bool,
        typer.Option(
            '--stop-on-failure',
            help='Stop processing files immediately if an error occurs with any file',
        ),
    ] = False,
):
    """
    Parse documents using one or more drivers.

    This command processes PDF documents and extracts their content in various formats.
    You can specify individual files or entire folders to process.

    Examples:

        # Parse a single file
        parxy parse document.pdf

        # Parse multiple files with specific driver
        parxy parse doc1.pdf doc2.pdf -d pymupdf

        # Parse all PDFs in a folder (non-recursive by default)
        parxy parse /path/to/folder

        # Parse all PDFs in a folder and subdirectories (recursive)
        parxy parse /path/to/folder --recursive

        # Parse with limited recursion depth (max 2 levels deep)
        parxy parse /path/to/folder --recursive --max-depth 2

        # Use multiple drivers
        parxy parse document.pdf -d pymupdf -d llamaparse

        # Output as JSON and show in console
        parxy parse document.pdf -m json --show
    """
    console.action('Parse files', space_after=False)
    # Collect all files
    files = collect_files(inputs, recursive=recursive, max_depth=max_depth)

    if not files:
        console.warning('No suitable files found to process.', panel=True)
        raise typer.Exit(1)

    # Use default driver if none specified
    if not drivers:
        drivers = [Parxy.default_driver()]  # Will use default driver

    # Convert output_dir to Path if specified
    output_path = Path(output_dir) if output_dir else None

    # Calculate total tasks
    total_tasks = len(files) * len(drivers)

    # Determine if we should use driver suffix (when multiple drivers are used)
    use_driver_suffix = len(drivers) > 1

    if use_driver_suffix:
        console.info(
            'You have specified more than one driver. Driver name will be added as suffix to the file name while saving.'
        )

    error_count = 0

    # Show info
    with console.shimmer(
        f'Processing {len(files)} file{"s" if len(files) > 1 else ""} with {len(drivers)} driver{"s" if len(drivers) > 1 else ""}...'
    ):
        # Process files with progress bar
        with console.progress('Processing documents') as progress:
            task = progress.add_task('', total=total_tasks)

            for file_path in files:
                for driver in drivers:
                    try:
                        output_file, page_count = process_file_with_driver(
                            file_path=file_path,
                            driver=driver,
                            level=level,
                            mode=mode,
                            output_dir=output_path,
                            show=show,
                            use_driver_suffix=use_driver_suffix,
                        )

                        # Update progress
                        console.print(
                            f'[faint]⎿ [/faint] {file_path.name} via {driver} to [success]{output_file}[/success] [faint]({page_count} pages)[/faint]'
                        )
                        progress.update(task, advance=1)

                    except Exception as e:
                        console.print(
                            f'[faint]⎿ [/faint] {file_path.name} via {driver} error. [error]{str(e)}[/error]'
                        )
                        progress.update(task, advance=1)
                        error_count += 1

                        if stop_on_failure:
                            console.newline()
                            console.info(
                                'Stopping due to error (--stop-on-failure flag is set)'
                            )
                            raise typer.Exit(1)

                        continue

            elapsed_time = format_timedelta(
                timedelta(seconds=max(0, progress.tasks[0].elapsed))
            )

    console.newline()
    if error_count == len(files):
        console.error('All files were not processed due to errors')
        return

    if error_count > 0:
        console.warning(
            f'Processed {len(files)} file{"s" if len(files) > 1 else ""} with warnings using {len(drivers)} driver{"s" if len(drivers) > 1 else ""}'
        )
        console.print(
            f'[faint]⎿ [/faint] [highlight]{error_count} files errored[/highlight]'
        )
        return

    console.success(
        f'Processed {len(files)} file{"s" if len(files) > 1 else ""} using {len(drivers)} driver{"s" if len(drivers) > 1 else ""} (took {elapsed_time})'
    )



import os
import time
from typing import Optional, List, Annotated

import typer

from parxy_core.facade import Parxy
from parxy_cli.models import Level
from parxy_cli.console.console import Console

app = typer.Typer()

console = Console()


@app.command()
def markdown(
    files: Annotated[
        List[str],
        typer.Argument(
            help='One or more files to parse',
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
        ),
    ],
    driver: Annotated[
        Optional[str],
        typer.Option(
            '--driver',
            '-d',
            help='Driver to use for parsing (default: pymupdf or PARXY_DEFAULT_DRIVER)',
        ),
    ] = None,
    level: Annotated[
        Level,
        typer.Option(
            '--level',
            '-l',
            help='Extraction level',
        ),
    ] = Level.BLOCK,
    env_file: Annotated[
        str,
        typer.Option(
            '--env',
            '-e',
            help='Path to .env file with configuration',
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
        ),
    ] = '.env',
    output_dir: Annotated[
        Optional[str],
        typer.Option(
            '--output',
            '-o',
            help='Directory to save markdown files. If not specified, output will be printed to console',
            dir_okay=True,
            file_okay=False,
        ),
    ] = None,
    combine: Annotated[
        bool,
        typer.Option(
            '--combine',
            '-c',
            help='Combine all documents into a single markdown file',
        ),
    ] = False,
):
    """Parse documents to Markdown."""
    try:
        # Create output directory if specified
        if output_dir:
            os.makedirs(output_dir, exist_ok=True)

        # For combined output
        combined_content = []

        # Process each file
        for file_path in files:
            try:
                with console.shimmer(f'Processing {file_path}...'):
                    # Parse the document
                    doc = Parxy.parse(
                        file=file_path,
                        level=level.value,
                        driver_name=driver,
                    )

                console.action(file_path)
                console.faint(f'{len(doc.pages)} pages extracted.')

                # Prepare markdown content
                file_info = f"""```yaml
file: "{file_path}"
pages: {len(doc.pages)}
```"""
                header = f'# {os.path.basename(file_path)}\n'
                content = doc.markdown()

                markdown_content = f'{file_info}\n{header}\n{content}'

                if output_dir and not combine:
                    # Generate output filename
                    base_name = os.path.splitext(os.path.basename(file_path))[0]
                    output_path = os.path.join(output_dir, f'{base_name}.md')

                    # Save to file
                    with open(output_path, 'w', encoding='utf-8') as f:
                        f.write(markdown_content)
                    console.success(f'Saved to: {output_path}.')

                elif not output_dir:
                    # Print to console
                    console.markdown(markdown_content)
                    console.rule()
                    console.newline()

                if combine:
                    combined_content.append(markdown_content)

            except Exception as e:
                console.error(f'Error processing {file_path}: {str(e)}')

        # Save combined content if requested
        if combine and output_dir and combined_content:
            output_path = os.path.join(output_dir, 'combined_output.md')
            with open(output_path, 'w', encoding='utf-8') as f:
                f.write('\n\n---\n\n'.join(combined_content))
            console.success(f'Combined output saved to: {output_path}')

    except Exception as e:
        console.error(f'Error: {str(e)}')
        raise typer.Exit()



import io
import requests
import validators

from typing import Self
from abc import ABC, abstractmethod
from logging import Logger

from parxy_core.models import Document
from parxy_core.exceptions import (
    FileNotFoundException,
    ParsingException,
    AuthenticationException,
)
from parxy_core.models.config import BaseConfig
from parxy_core.logging import create_null_logger


class Driver(ABC):
    """Define a document processing driver.

    This class is intended to be abstract to serve as the starting point for implementing your own document processing

    Note to implementers:
    - Subclasses must implement the `_handle` method and declare the supported extraction levels.
    - If you need to initialize specific driver state override the _initialize method, as it will be called during class instantiation

    Attributes
    ----------
    supported_levels : list of str
        The list of supported extraction levels (e.g., `page`, `block`, `line`, etc.).

    _config : BaseConfig
        The configuration dictionary with specific parameters for initializing the document processing driver.

    _logger : Logger
        The logger instance.
    """

    supported_levels: list[str] = ['page', 'block']

    _config: BaseConfig

    _logger: Logger

    def __new__(cls, config: BaseConfig = None, logger: Logger = None):
        instance = super().__new__(cls)
        instance.__init__(config=config, logger=logger)
        return instance

    def __init__(self, config: BaseConfig = None, logger: Logger = None):
        self._config = config

        if logger is None:
            logger = create_null_logger(name=f'parxy.{self.__class__.__name__}')

        self._logger = logger
        self._initialize_driver()

    def parse(
        self,
        file: str | io.BytesIO | bytes,
        level: str = 'block',
        **kwargs,
    ) -> Document:
        """Parse a file

        Parameters
        ----
        file : str | io.BytesIO | bytes
            Path, URL or stream of the file to parse.
        level : str, optional
            Desired extraction level. Must be one of `supported_levels`. Default is `block`.
        **kwargs : dict
            Additional keyword arguments that can be passed to the specific driver.

        Returns
        -------
        Document
            A `Document` object representing the content.

        Throws
        -------
        ValueError
            If input is invalid, e.g. invalid URL, level

        FileNotFoundException
            If the file or URL is not accessible

        AuthenticationException
            In case authentication fails to third party service

        ParsingException
            For all unhandled exceptions during document processing
        """

        self._logger.debug(f'Parsing file using {self.__class__.__name__}', file)

        self._validate_level(level)

        try:
            document = self._handle(file=file, level=level, **kwargs)

            return document

        except Exception as ex:
            self._logger.error(
                f'Error while parsing file [{str(file)}]: {ex.message if hasattr(ex, "message") else str(ex)}',
                file,
                self.__class__.__name__,
                exc_info=True,
            )

            if isinstance(ex, FileNotFoundError):
                raise FileNotFoundException(ex, self.__class__) from ex
            elif isinstance(
                ex, (FileNotFoundException, AuthenticationException, ParsingException)
            ):
                raise ex
            raise ParsingException(ex, self.__class__) from ex

    def _initialize_driver(self) -> Self:
        """Initialize driver internal logic. It is called automatically during class initialization"""
        return self

    @abstractmethod
    def _handle(
        self,
        file: str | io.BytesIO | bytes,
        level: str = 'block',
        **kwargs,
    ) -> Document:
        pass

    def _validate_level(self, level: str):
        if level not in self.supported_levels:
            raise ValueError(
                f'The level is not supported. Expected [{self.supported_levels}], received [{level}].'
            )

    @classmethod
    def get_stream_from_url(cls, url: str) -> io.BytesIO:
        if validators.url(url) is False:
            raise ValueError(f'The given input [`{url}`] is not a valid URL.')

        response = requests.get(url, stream=True)

        try:
            response.raise_for_status()
        except requests.HTTPError as hex:
            if hex.response.status_code in (401, 403):
                raise AuthenticationException(
                    f'Authentication error while downloading [{url}].', cls.__name__
                ) from hex
            elif hex.response.status_code == 404:
                raise FileNotFoundException(
                    f'File not found while downloading [{url}].', cls.__name__
                ) from hex

        file_input = io.BytesIO(response.content)
        file_input.name = url.split('/')[
            -1
        ]  # TODO: improve as some urls have accessible filename, others don't
        return file_input



"""
Parxy Console class

Provides methods for styled output, progress bars, and spinners.

The theme uses the warm, inky Flexoki color scheme by Steph Ango https://stephango.com/flexoki
"""

import os
import sys
from rich.console import (
    ConsoleOptions,
    RenderResult,
    Console as RichConsole,
)
from rich.theme import Theme
from rich.progress import (
    Progress,
    TextColumn,
    BarColumn,
    TaskProgressColumn,
    MofNCompleteColumn,
    TimeElapsedColumn,
)
from rich.markdown import Markdown, Heading, TextElement, MarkdownContext
from rich.panel import Panel
from rich.live import Live
from rich.text import Text
from rich.style import Style
from rich.padding import Padding
from contextlib import contextmanager
from markdown_it.token import Token

from parxy_core.models.config import ParxyConfig

# Flexoki color palette (dark theme - 400 series)
COLORS_DARK = {
    # Base colors
    'bg': '#1C1B1A',
    'bg_2': '#282726',
    'ui': '#343331',
    'ui_2': '#403E3C',
    'ui_3': '#575653',
    'tx_3': '#B7B5AC',
    'tx_2': '#CECDC3',
    'tx': '#E6E4D9',
    # Accent colors (400 series for dark theme)
    'black': '#100F0F',
    'white': '#FFFCF0',
    'red': '#D14D41',
    'orange': '#DA702C',
    'yellow': '#D0A215',
    'green': '#879A39',
    'cyan': '#3AA99F',
    'blue': '#4385BE',
    'purple': '#8B7EC8',
    'magenta': '#CE5D97',
}

# Flexoki color palette (light theme - 600 series)
COLORS_LIGHT = {
    # Base colors
    'bg': '#FFFCF0',
    'bg_2': '#F2F0E5',
    'ui': '#E6E4D9',
    'ui_2': '#DAD8CE',
    'ui_3': '#B7B5AC',
    'tx_3': '#6F6E69',
    'tx_2': '#403E3C',
    'tx': '#100F0F',
    # Accent colors (600 series for light theme)
    'black': '#100F0F',
    'white': '#FFFCF0',
    'red': '#AF3029',
    'orange': '#BC5215',
    'yellow': '#AD8301',
    'green': '#66800B',
    'cyan': '#24837B',
    'blue': '#205EA6',
    'purple': '#5E409D',
    'magenta': '#A02F6F',
}


class MarkdownHeading(TextElement):
    """A heading."""

    @classmethod
    def create(cls, markdown: Markdown, token: Token) -> Heading:
        return cls(token)

    def on_enter(self, context: MarkdownContext) -> None:
        self.text = Text()
        context.enter_style(self.style_name)

    def __init__(self, token: Token) -> None:
        self.token = token
        self.style_name = f'markdown.{token.tag}'
        super().__init__()

    def __rich_console__(
        self, console: RichConsole, options: ConsoleOptions
    ) -> RenderResult:
        yield (
            Text('#' * int(self.token.tag[1]), style=self.style_name)
            .append(' ')
            .append_text(self.text)
        )


Markdown.elements['heading_open'] = MarkdownHeading


class Shimmer:
    """
    A renderable that creates a shimmering text effect.
    The shimmer is a wave that dims characters as it passes over them.

    This is a stateful renderable that advances its animation on each render,
    similar to Rich's Spinner class.
    """

    def __init__(
        self,
        text: str,
        normal_color: str,
        dim_color: str,
        mid_color: str,
        speed: float = 1.0,
    ):
        """
        Initialize the Shimmer renderable.

        Args:
            text: The text to shimmer
            normal_color: Color for normal (bright) characters
            dim_color: Color for the dimmest characters (center of wave)
            mid_color: Color for slightly dimmed characters (adjacent to center)
            speed: Animation speed multiplier (higher = faster)
        """
        self.text = text
        self.normal_color = normal_color
        self.dim_color = dim_color
        self.mid_color = mid_color
        self.speed = speed

        # Animation state
        self.position = 0
        self.direction = 1  # 1 for forward, -1 for backward
        self._frame_count = 0

    def __rich_console__(self, console, options):
        """Render the shimmer effect for the current frame."""
        # Advance animation based on speed
        # Update position every N frames based on speed
        frames_per_step = max(1, int(1.0 / self.speed))

        if self._frame_count % frames_per_step == 0:
            self.position += self.direction

            # Bounce at edges
            if self.position >= len(self.text) - 1:
                self.direction = -1
            elif self.position <= 0:
                self.direction = 1

        self._frame_count += 1

        # Create the shimmered text
        text = Text()
        for i, char in enumerate(self.text):
            distance = abs(i - self.position)

            if distance == 0:
                # Center of wave - most dimmed
                text.append(char, style=self.dim_color)
            elif distance == 1:
                # Adjacent to center - slightly dimmed
                text.append(char, style=self.mid_color)
            else:
                # Normal brightness
                text.append(char, style=self.normal_color)

        yield text

    def __rich_measure__(self, console, options):
        """Return the width of the shimmer text."""
        from rich.measure import Measurement

        text_length = len(self.text)
        return Measurement(text_length, text_length)


class Console:
    """
    A themed console wrapper.
    Provides methods for styled output, progress bars, and spinners.
    Tries to automatically detects terminal background and uses appropriate theme.
    """

    @staticmethod
    def detect_terminal_background(config: ParxyConfig = None):
        """
        Detect if the terminal has a light or dark background.
        Returns 'dark' or 'light'.

        Detection methods:
        1. Check ParxyConfig theme setting
        2. Check COLORFGBG environment variable
        3. Check TERM_PROGRAM for known terminals
        4. Default to 'dark' if uncertain

        Args:
            config: Optional ParxyConfig instance. If provided and theme is set, uses that value.
        """
        # Method 1: Check ParxyConfig theme setting (highest priority)
        if config is not None and config.theme is not None:
            return config.theme

        # Method 2: Check COLORFGBG environment variable
        # Format is typically "foreground;background" where background color:
        # 0-6 or 8 = dark background, 7 or 15 = light background
        colorfgbg = os.environ.get('COLORFGBG', '')
        if colorfgbg:
            parts = colorfgbg.split(';')
            if len(parts) >= 2:
                try:
                    bg_color = int(parts[-1])
                    # Background colors 0-6, 8 are dark; 7, 15 are light
                    if bg_color in (7, 15):
                        return 'light'
                    elif bg_color in (0, 1, 2, 3, 4, 5, 6, 8):
                        return 'dark'
                except ValueError:
                    pass

        # Method 3: Check for Windows Terminal light theme
        if sys.platform == 'win32':
            os.environ.get('WT_PROFILE_ID', '')
            # Windows Terminal doesn't expose theme directly, but we can check registry
            # For now, we'll default based on common settings
            pass

        # Method 4: Check TERM_PROGRAM for known defaults
        term_program = os.environ.get('TERM_PROGRAM', '').lower()
        if term_program in ('apple_terminal', 'iterm.app'):
            # These often default to light themes
            # But we can't be certain, so we won't make assumptions
            pass

        # Default to dark theme (most terminal defaults)
        return 'dark'

    def __init__(self, theme_mode=None, config: ParxyConfig = None):
        """
        Initialize the console.

        Args:
            theme_mode: Optional theme mode ('light' or 'dark').
                       If None, auto-detects based on config or terminal background.
            config: Optional ParxyConfig instance for loading theme from configuration.
        """
        # Load config if not provided
        if config is None:
            config = ParxyConfig()

        # Detect or use specified theme
        if theme_mode is None:
            theme_mode = self.detect_terminal_background(config)

        # Select appropriate color palette
        self.theme_mode = theme_mode
        self.COLORS = COLORS_LIGHT if theme_mode == 'light' else COLORS_DARK

        # Build theme with selected colors
        # Extend https://github.com/Textualize/rich/blob/master/rich/default_styles.py
        self.theme = Theme(
            {
                # Base text styles
                'default': f'{self.COLORS["tx"]}',
                'muted': f'{self.COLORS["tx_2"]}',
                'faint': f'{self.COLORS["tx_3"]}',
                'black': f'{self.COLORS["black"]}',
                'red': f'{self.COLORS["red"]}',
                'orange': f'{self.COLORS["orange"]}',
                'yellow': f'{self.COLORS["yellow"]}',
                'green': f'{self.COLORS["green"]}',
                'cyan': f'{self.COLORS["cyan"]}',
                'blue': f'{self.COLORS["blue"]}',
                'purple': f'{self.COLORS["purple"]}',
                'magenta': f'{self.COLORS["magenta"]}',
                'white': f'{self.COLORS["white"]}',
                # Message types
                'success': f'bold {self.COLORS["green"]}',
                'info': f'{self.COLORS["cyan"]}',
                'warning': f'bold {self.COLORS["orange"]}',
                'error': f'bold {self.COLORS["red"]}',
                # Semantic colors
                'highlight': f'bold {self.COLORS["yellow"]}',
                'link': f'underline {self.COLORS["blue"]}',
                'repr.number': Style(
                    color=self.COLORS['blue'], bold=True, italic=False
                ),
                'repr.number_complex': Style(
                    color=self.COLORS['blue'], bold=True, italic=False
                ),
                # Progress/spinner colors
                'bar.complete': f'{self.COLORS["blue"]}',
                'bar.finished': f'{self.COLORS["blue"]}',
                'bar.pulse': f'{self.COLORS["blue"]}',
                'progress.description': f'{self.COLORS["tx_2"]}',
                'progress.percentage': f'{self.COLORS["tx_2"]}',
                'progress.filesize': f'{self.COLORS["tx_2"]}',
                'progress.filesize.total': f'{self.COLORS["tx_2"]}',
                'progress.download': f'{self.COLORS["tx_2"]}',
                'progress.elapsed': f'{self.COLORS["tx_2"]}',
                'progress.remaining': f'{self.COLORS["tx_2"]}',
                'progress.data.speed': f'{self.COLORS["tx_2"]}',
                'progress.spinner': f'{self.COLORS["tx_3"]}',
                'status.spinner': f'{self.COLORS["tx_3"]}',
                # Markdown styles
                'markdown.h1': Style(bold=True, color=self.COLORS['magenta']),
                'markdown.h2': Style(bold=True, color=self.COLORS['magenta']),
                'markdown.h3': Style(bold=True, color=self.COLORS['magenta']),
                'markdown.h4': Style(bold=True, color=self.COLORS['magenta']),
                'markdown.h5': Style(bold=True, color=self.COLORS['magenta']),
                'markdown.h6': Style(bold=True, color=self.COLORS['magenta']),
                'markdown.h7': Style(bold=True, color=self.COLORS['magenta']),
                'markdown.item.bullet': f'{self.COLORS["tx_3"]}',
                'markdown.item.number': f'{self.COLORS["tx_3"]}',
                'markdown.link': Style(color=self.COLORS['blue']),
                'markdown.link_url': Style(color=self.COLORS['blue'], underline=True),
                'markdown.hr': Style(color=self.COLORS['ui_2']),
                'markdown.block_quote': Style(
                    color=self.COLORS['tx_2'],
                    bgcolor=self.COLORS['ui'],
                    bold=False,
                ),
                # The blockquote bar on the left
                'markdown.blockquote.border': Style(
                    color=self.COLORS['cyan'], bold=True
                ),
                'markdown.code_block': Style(
                    bgcolor=self.COLORS['ui'], color=self.COLORS['tx']
                ),
                'code': Style(bgcolor=self.COLORS['ui'], color=self.COLORS['tx']),
                'markdown.code': Style(
                    bgcolor=self.COLORS['ui'], color=self.COLORS['tx']
                ),
                # ISO 8601 styles
                'iso8601.date': Style(bold=True),
                'iso8601.time': Style(bold=True),
                'iso8601.timezone': Style(bold=True),
            }
        )

        self.console = RichConsole(theme=self.theme)

    def print(self, *args, style=None, **kwargs):
        """Print with optional style."""
        self.console.print(*args, style=style, **kwargs)

    def _icon_and_text(
        self,
        message: str,
        icon: str = '✓',
        icon_style: str = 'default',
        padding: int = 1,
    ):
        from rich.table import Table

        grid = Table.grid(padding=(0, padding), expand=False)
        grid.add_column(
            width=1,
        )
        grid.add_column()

        grid.add_row(Text(icon, style=icon_style), Text(message))

        return grid

    def success(self, message: str, prefix: str = '✓', panel: bool = False):
        """Print a success message."""
        # prefix_text = Text(prefix, style='success')
        # padded_prefix = Padding(prefix_text, (0, 1))
        formatted = self._icon_and_text(
            message=message, icon=prefix, icon_style='success'
        )
        self.print(formatted) if not panel else self.panel(
            formatted, border_style='success'
        )

    def info(self, message: str, prefix: str = 'ℹ', panel: bool = False):
        """Print an info message."""
        formatted = self._icon_and_text(message=message, icon=prefix, icon_style='info')
        self.print(formatted) if not panel else self.panel(
            formatted, border_style='info'
        )

    def warning(self, message: str, prefix: str = '⚠', panel: bool = False):
        """Print a warning message."""
        formatted = self._icon_and_text(
            message=message, icon=prefix, icon_style='warning'
        )
        self.print(formatted) if not panel else self.panel(
            formatted, border_style='warning'
        )

    def error(self, message: str, prefix: str = '✗', panel: bool = False):
        """Print an error message."""
        formatted = self._icon_and_text(
            message=message, icon=prefix, icon_style='error'
        )
        self.print(formatted) if not panel else self.panel(
            formatted, border_style='error'
        )

    def muted(self, message: str):
        """Print muted text."""
        self.print(message, style='muted')

    def faint(self, message: str):
        """Print faint text."""
        self.print(message, style='faint')

    def highlight(self, message: str):
        """Print highlighted text."""
        self.print(message, style='highlight')

    def parxy(self):
        """Print Parxy and its tagline."""
        self.action('[bold]Parxy[/bold]', style='blue')
        self.print('[faint][italic]Every document matters.[/italic][/faint]')
        self.newline()

    def action(
        self,
        message: str,
        style: str = 'faint',
        space_before: bool = False,
        space_after: bool = True,
    ):
        """Print a highlighted action."""
        if space_before:
            self.newline()
        self.print(f'[{style}]▣[/{style}] {message}')
        if space_after:
            self.newline()

    def markdown(self, content: str, code_theme: str = 'monokai'):
        """
        Render markdown content.

        Args:
            content: Markdown content to render
            code_theme: Syntax highlighting theme for code blocks (default: 'monokai')
        """
        md = Markdown(
            content,
            code_theme=code_theme,
            hyperlinks=True,
            inline_code_lexer=None,
            inline_code_theme=code_theme,
        )
        self.console.print(md)

    def panel(
        self,
        content: str,
        title: str = None,
        style: str = 'default',
        border_style: str = None,
    ):
        """Display content in a panel."""
        border_color = border_style or self.COLORS['ui_2']
        panel = Panel(
            content,
            title=title,
            border_style=border_color,
            title_align='left',
            style=style,
        )
        self.console.print(panel)

    def quote(self, content: str, expand: bool = False):
        """
        Display a blockquote-style callout with cyan left border and background.
        Similar to markdown blockquotes but with enhanced styling.

        Args:
            content: The quote content (can be markdown)
            title: Optional title for the quote
        """
        from rich.table import Table

        # Create a table with a cyan left border and background
        table = Table.grid(padding=(0, 1), expand=expand)
        table.add_column(
            style=Style(
                color=self.COLORS['cyan'], bgcolor=self.COLORS['ui'], bold=True
            ),
            width=1,
        )
        table.add_column(
            style=Style(
                color=self.COLORS['tx_2'],
                bgcolor=self.COLORS['ui'],
            )
        )

        # Split content into lines for proper rendering
        lines = content.strip().split('\n')
        for line in lines:
            table.add_row('▌', line if line else ' ')

        # Add padding around the quote
        self.console.print(Padding(table, (1, 0)))

    def separator(self, title: str = None, style: str = None):
        """Print a horizontal separator."""
        rule_style = style or self.COLORS['ui_2']
        self.console.rule(title, style=rule_style, align='left')

    @contextmanager
    def progress(self, description: str = 'Working...'):
        """
        Context manager for a progress bar.

        Usage:
            with console.progress("Processing files") as progress:
                task = progress.add_task("", total=100)
                for i in range(100):
                    progress.update(task, advance=1)
        """
        progress = Progress(
            TextColumn('[progress.description]{task.description}'),
            BarColumn(),
            TaskProgressColumn(),
            MofNCompleteColumn(),
            TimeElapsedColumn(),
            console=self.console,
            transient=True,
        )

        with progress:
            yield progress

    @contextmanager
    def spinner(self, message: str = 'Loading...'):
        """
        Context manager for a spinner.

        Usage:
            with console.spinner("Fetching data..."):
                # do work
                time.sleep(2)
        """
        with self.console.status(
            f'[{self.COLORS["cyan"]}]{message}[/{self.COLORS["cyan"]}]',
            spinner='dots',
            spinner_style='bar.pulse',
        ):
            yield

    @contextmanager
    def shimmer(self, message: str = 'Loading...', speed: float = 1.0):
        """
        Context manager for a text shimmering effect.
        Creates a wave of dimming through individual characters.
        The shimmer is automatically cleared when the task completes.

        Args:
            message: The text to display with shimmer effect
            speed: Animation speed multiplier (default 1.0, higher = faster)

        Usage:
            with console.shimmer("Processing data..."):
                # do work
                time.sleep(2)
        """
        # Create Shimmer renderable with theme colors
        shimmer = Shimmer(
            text=message,
            normal_color=self.COLORS['tx'],
            dim_color=self.COLORS['tx_3']
            if self.theme_mode == 'dark'
            else self.COLORS['ui_3'],
            mid_color=self.COLORS['tx_2']
            if self.theme_mode == 'dark'
            else self.COLORS['ui_2'],
            speed=speed,
        )

        # Use Live display with transient=True to clear the shimmer when done
        with Live(
            shimmer,
            console=self.console,
            refresh_per_second=20,
            transient=True,
        ):
            yield

    def clear(self):
        """Clear the console."""
        self.console.clear()

    def newline(self, count: int = 1):
        """Print newlines."""
        self.console.print('\n' * (count - 1))

    def get_theme_mode(self):
        """Get the current theme mode ('light' or 'dark')."""
        return self.theme_mode

    @contextmanager
    def pager(self, styles: bool = False):
        """
        Context manager for paginated output.

        Usage:
            with console.pager():
                console.print(large_content)

        Args:
            styles: If True, ANSI styles will be preserved in the pager (default: False)
        """
        with self.console.pager(styles=styles):
            yield


# Example usage
if __name__ == '__main__':
    import time

    # Console automatically detects terminal background
    console = Console()

    console.markdown('# Parxy Cli Theme Demo')

    console.info(f'Using {console.get_theme_mode()} theme')
    console.newline()

    # Basic messages
    console.markdown('# Message Types')

    console.success('Operation completed successfully!')
    console.info('This is an informational message')
    console.warning('This is a warning message')
    console.error('This is an error message')
    console.newline()

    console.success('Operation completed successfully!', panel=True)
    console.info('This is an informational message', panel=True)
    console.warning('This is a warning message', panel=True)
    console.error('This is an error message', panel=True)
    console.newline()

    # Different text styles
    console.markdown('# Text styles')
    console.print('This is [highlight]highlighted[/highlight] text')
    console.print('This has [code]inline code[/code] text')
    console.muted('This is muted text')
    console.faint('This is faint text')
    console.print(
        '[bold]bold[/bold], [em]italic[/em] and [underline]underline[/underline]'
    )
    console.newline()

    console.markdown('# Link')
    console.print(
        '[link=https://github.com/OneOffTech/parxy]This is a clickable link[/link]'
    )
    console.newline()

    # Markdown
    console.markdown("""
# Parxy Console

This console uses the [**Flexoki**](https://stephango.com/flexoki) color scheme for a warm, inky feel.

1. Auto-detects light/dark terminal background
2. Themed messages (success, info, warning, error)
3. Progress bars and spinners
4. Shimmer effect for long operations

- a bullet
- in a bullet list

## All headings
                     
### heading 3
#### heading 4
##### heading 5
###### heading 6

   
## Example Code

```python
def hello_world():
    print("Hello, Parxy!")
```

```
def hello_world():
    print("Hello, Parxy!")
```
                     
`Single line code block example`
                     

> Block quote
    """)
    console.newline()

    console.markdown('# Panels')
    # Panel
    console.panel(
        'This is content inside a panel with a nice border',
        title='Panel Example',
    )
    console.newline()

    console.panel(
        'Panel without a title',
    )
    console.newline()

    # Quote / Callout
    console.quote(
        'This is a beautifully styled blockquote with a cyan left border\n'
        'and a subtle background, perfect for highlighting important information.\n\n'
        'It supports multiple lines and looks great!'
    )
    console.newline()

    console.markdown('# Dividers')

    console.markdown('----')

    console.separator()

    console.separator('Separator with title')

    console.newline()
    console.markdown('# Progress')
    # Progress bar
    with console.progress('Processing items') as progress:
        task = progress.add_task('', total=50)
        for i in range(50):
            time.sleep(0.02)
            progress.update(task, advance=1)

    console.newline()

    # Spinner
    with console.spinner('Fetching data from API...'):
        time.sleep(2)

    console.newline()

    # Shimmer
    with console.shimmer('Fetching data from API...'):
        time.sleep(3)

    console.success('All demonstrations complete!')



class UnsupportedFormatException(Exception):
    """Exception raised for file format not supported.

    This exception is raised when a file is of a format not supported by the parsing service.

    Attributes
    ----------
    message : str
        Explanation error
    service : str
        Name of the service that failed (e.g., 'LlamaParse', 'PDFAct')
    details : dict, optional
        Additional details about the error, such as response data or error codes

    Example
    ---------
    try:
        # API call fails
        raise UnsupportedFormatException(
            message="Format [text/markdown] not supported.",
            service="LlamaParse",
            details={"error_code": "ERR001", "response": {"detail": "Unknown format"}}
        )
    except UnsupportedFormatException as e:
        print(e)  # Will print: "Unsupported format for LlamaParse: Format [text/markdown] not supported.\nDetails: {...}"
    """

    def __init__(self, message: str, service: str, details: dict = None):
        """Initialize the exception.

        Parameters
        ----------
        message : str
            Human-readable error message
        service : str
            Name of the service that failed
        details : dict, optional
            Additional error details, by default None
        """
        self.message = message
        self.service = service
        self.details = details or {}
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return a string representation of the error.

        Returns
        -------
        str
            Formatted error message including service name and details
        """
        base_message = f'Unsupported format for {self.service}: {self.message}'
        if self.details:
            return f'{base_message}\nDetails: {self.details}'
        return base_message



# Use an explicit re-export https://github.com/astral-sh/ruff/issues/5697#issuecomment-1631647211

from parxy_core.models.models import (
    BoundingBox as BoundingBox,
    Style as Style,
    Character as Character,
    Span as Span,
    Line as Line,
    Block as Block,
    ImageBlock as ImageBlock,
    TableBlock as TableBlock,
    TextBlock as TextBlock,
    Page as Page,
    Metadata as Metadata,
    Document as Document,
    # estimate_lines_from_block,
    HierarchyLevel as HierarchyLevel,
)

from parxy_core.models.config import (
    ParxyConfig as ParxyConfig,
    PdfActConfig as PdfActConfig,
    LlamaParseConfig as LlamaParseConfig,
    LandingAIConfig as LandingAIConfig,
    LlmWhispererConfig as LlmWhispererConfig,
    UnstructuredLocalConfig as UnstructuredLocalConfig,
)
